/*
	PROforma, PROGS Font & Raster Manager

	Canon BJ-10 & compatible COLOUR bubblejet printer driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 1 May 1996, based on Epson 24 pin printer driver and BJ-10
						printer driver for PROforma v1.06
	27.06.1996 should now work on all Canon BJ printers
	07.04.1997 adjustments for colour
	29.04.1997 colour selection was wrong
	16.12.1997 - better handling of yskip
			   - no auto line spacing etc
	06.01.1997 - bug fixed where black plane was reprint of magenta
			   - slight reduction of code size
			   - last line now handled properly
	07.02.1998 much improved, should now work 100%
	10.03.1998 buffers now properly cleared at start
*/

#include <io.h>
#include <mem.h>
#include "PFmodule.h"

typedef struct _DriverInfo {
	Channel channel;		/* the output channel for this driver */
	int bit;				/* next bit for transform */
	int xskip, yskip;		/* skip values to end up at the topleft */
	char *buffer;
} DriverInfo;


Bbox PrintableArea={
	double2pt(577.8), short2pt(780),/* size */
	double2pt(8.6), short2pt(24)	/* origin */
};

char PrintDevice[IO_MAXFULLNAME]="pard";

Error Define(char *name, char *value)
{
	double x,y;

	if (STRSameCD(name,"DEFAULT-DEVICE"))
	{
		STRCopySafe(value,PrintDevice,IO_MAXFULLNAME);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-SIZE"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xsiz=double2pt(x);
		PrintableArea.ysiz=double2pt(y);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-ORIGIN"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xorg=double2pt(x);
		PrintableArea.yorg=double2pt(y);
		return ERR_OK;
	}
	return ERR_IPAR;
}

Error InitInit(Gstate gstate, char *device, PBbox *area)
{
	Error err;
	DriverInfo *di;
	DRIVdriver *prt=gstate->printer;

	err=MEMAllocate(sizeof(DriverInfo),&di);
	if (err) return err;

	if (device && *device)
		err=IOOpen(device,OPEN_OVER,&di->channel);
	else
		err=IOOpen(PrintDevice,OPEN_OVER,&di->channel);
	if (err) { MEMRelease(di); return err; };

	gstate->PrinterExtra=di;
	gstate->todev.x = pt_one * prt->xdpi / 72;
	gstate->todev.y = pt_one * prt->ydpi / 72;
	gstate->ppp.x = pt_one * 72 / prt->xdpi;
	gstate->ppp.y = pt_one * 72 / prt->ydpi;

	area->xsiz=pt2rshort(fixmul(PrintableArea.xsiz,gstate->todev.x));
	area->ysiz=pt2rshort(fixmul(PrintableArea.ysiz,gstate->todev.y));
	area->xorg=pt2rshort(fixmul(PrintableArea.xorg,gstate->todev.x));
	area->yorg=pt2rshort(fixmul(PrintableArea.yorg,gstate->todev.y));

	return err;
}

Error InitDone(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size linc;

	/* complete the initialisation of the DriverInfo */
	di->buffer=(char *)gstate->raster->base;
	gstate->raster->base+=gstate->raster->linc*(48+24);
		/* extra 24 for buffer */

	di->xskip=pt2rshort(gstate->pagebbox.xporg-fixmul(PrintableArea.xorg,
							   gstate->todev.x));
	di->yskip=pt2rshort(gstate->pagebbox.yporg-fixmul(PrintableArea.yorg,
							   gstate->todev.y));

	return ERR_OK;
}

Error BufferSize(Gstate gstate, int xsiz, int ysiz, Size *size)
{
	DriverInfo *di=gstate->PrinterExtra;
	return gstate->bitmap->BufferSize(xsiz,ysiz+48+48,size);
		/* extra 24 lines at each side of each plane as buffer */
		/* this should be useless, but otherwise, I don't know what */
		/* could cause the overflow */
}

Error CleanUp(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di)
	{
		IOClose(di->channel);
		MEMRelease(gstate->PrinterExtra);
	}
	return ERR_OK;
}

Error ColourNearest(Gstate gstate, ColourCMYK *colour, int *result)
{
	*result=0;						/* assume white */
	if (colour->cyan>pt_hundred/2)
		*result+=8;
	if (colour->magenta>pt_hundred/2)
		*result+=4;
	if (colour->yellow>pt_hundred/2)
		*result+=2;
	if (colour->black>pt_hundred/2)
		*result+=1;
	return ERR_OK;
}

Error ColourDelta(Gstate gstate, ColourCMYK *colour, ColourCMYK *delta)
{
	*delta=*colour;
	if (colour->cyan>pt_hundred/2)
		delta->cyan-=pt_hundred;
	if (colour->magenta>pt_hundred/2)
		delta->magenta-=pt_hundred;
	if (colour->yellow>pt_hundred/2)
		delta->yellow-=pt_hundred;
	if (colour->black>pt_hundred/2)
		delta->black-=pt_hundred;
	return ERR_OK;
}

/* !!!!!!!!!!! code to send data to printer !!!!!!!!!!! */

/* some macros for generating output */
#define printb(s,b) IOPutRawBytes(chid,s,b,NULL)
#define prints(s)	printb(strlen(s),s)
#define printr(s)	printb(sizeof(s)-1,s);
#define printa(s)	printb(sizeof(s),s);
#define printc(c)	{ char s=c; IOPutRawBytes(chid,1,&s,NULL); }

/* routine to fill in a scan line in the transposed data */
static const char setbit[8] = { (char)0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static void transpose(char *data, char *dest, int bit, int skip, int linelen)
{
	char mask=setbit[bit&7];
	int i,j;

	dest+=bit/8;
	for (j=linelen ; j ; j--)
	{
		char test=*data++;
		for (i=0; i<8; i++)
		{
			if (test & setbit[i]) *dest|=mask;
			dest+=skip;
		}
	}
}

static const char seq_init[]   ={ 27,91,75,4,0,4,37,0,0,
								  27,91,92,4,0,0,0,104,1,
								  27,51,48 };
static const char seq_yellow[] ={ 27,91,77,1,0,6 };
static const char seq_cyan[]   ={ 27,91,77,1,0,5 };
static const char seq_magenta[]={ 27,91,77,1,0,3 };
static const char seq_black[]  ={ 27,91,77,1,0,0 };
static const char seq_bitmap[] ={ 28,67,66,4,4 };

/* routine to send a line to the printer */
static void printline(Gstate gstate, char *buff, int count)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int xskip=di->xskip/3;
	int xcount=(count+6-1)/6;

	/* make sure x position is correct */
	if (xskip && xcount)
	{
		printr("\033d");
		printc(xskip & 0xff);
		printc(xskip >> 8);
	}

	/* print the actual data */
	if (xcount)
	{
		printa(seq_bitmap);
		printc(xcount & 0xff);
		printc(xcount >> 8);
		printb(xcount*6,buff);
	}
	printr("\r");
}

Error InitPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int yskip=di->yskip,i;
	int bsiz=gstate->raster->linc*48;
	int pinc=gstate->raster->pinc;

	/* set graphics line spacing unit and auto line feed */
	printa(seq_init);

	/* Initialize the line skip */
	MEMClear(bsiz,di->buffer);
	MEMClear(bsiz,di->buffer+pinc);
	MEMClear(bsiz,di->buffer+pinc*2);
	MEMClear(bsiz,di->buffer+pinc*3);
	for (i=yskip/48 ; i ; i--) printr("\n");
	printr("\r");
	di->bit=yskip%48;

	return ERR_OK;
}

Error PrintAllPlanes(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int bsiz=gstate->raster->linc*48;
	int pinc=gstate->raster->pinc;
	char *buff, *buff_end;

#define COLOUR(count,colour) do {                       \
	buff=di->buffer+pinc*count; 						\
	buff_end=buff+bsiz; 								\
	while (buff_end>buff && !buff_end[-1]) buff_end--;	\
	if (buff!=buff_end) 								\
	{                                                   \
		printa(seq_##colour);							\
		printline(gstate, buff, buff_end-buff); 		\
	} } while (0)

	COLOUR(2,yellow);
	COLOUR(0,cyan);
	COLOUR(1,magenta);
	COLOUR(3,black);

	/* go to next line */
	printr("\n");
}

Error ShowPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;
	int lsiz=gstate->raster->linc;
	int bsiz=lsiz*48;
	int pinc=gstate->raster->pinc;
	char *data=(char *)gstate->raster->base;
	int i;

	for(i=gstate->pagebbox.ysiz ; i ; i--, data+=lsiz)
	{
		/* clear the buffer if necessary */
		if (di->bit==0)
		{
			MEMClear(bsiz,di->buffer);
			MEMClear(bsiz,di->buffer+pinc);
			MEMClear(bsiz,di->buffer+pinc*2);
			MEMClear(bsiz,di->buffer+pinc*3);
		}

		/* Add this scan line in the buffer (byteline) */
		/* C */ transpose(data		 ,di->buffer	   ,di->bit,6,lsiz);
		/* M */ transpose(data+pinc  ,di->buffer+pinc  ,di->bit,6,lsiz);
		/* Y */ transpose(data+pinc*2,di->buffer+pinc*2,di->bit,6,lsiz);
		/* K */ transpose(data+pinc*3,di->buffer+pinc*3,di->bit,6,lsiz);

		di->bit++;
		if (di->bit==48)
		{
			PrintAllPlanes(gstate);
			di->bit=0;
		}
	}
	return ERR_OK;
}

Error StopPage(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Channel chid=di->channel;

	if (di->bit)
		PrintAllPlanes(gstate);
	/* restore printing in black */
	printa(seq_black);
	printr("\f");
	return ERR_OK;
}


/* !!!!!!!!!!! printer driver definition !!!!!!!!!!! */

static Error nimp() 	{ return ERR_NIMP; }
static Error nothing()	{ return ERR_OK; }

Error InitDriver();

Error SetCopies(Gstate, int count);
Error SetDisplayMode(Gstate, int mode);
Error WindowMove(Gstate, int dx, int dy);
Error BufferUpdate(Gstate);
Error ColourAdjust(Gstate, ColourCMYK *);
Error GetFullPageBbox(Gstate, Bbox *box);
Error DefaultDevice(Gstate, char *device);

#define InitDriver		nothing
#define WindowMove		nothing
#define ColourAdjust	nothing
#define SetDisplayMode	nimp
#define GetFullPageBbox nimp
#define DefaultDevice	nimp
#define SetCopies		nimp
#define BufferUpdate	(Error (*)(Gstate))nimp
#define BufferSave		(Error (*)(Gstate))nimp
#define BufferRestore	(Error (*)(Gstate))nimp
#define Handle			(Error (*)(int,...))nimp

DRIVdriver init = {
	NULL, PF_PRINTERDRIVER,
	"Canon BJ colour, 360dpi",		/* name */
	360, 360,					/* xdpi, ydpi */
	{                           /* mono bitmap */
		4, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_SEPARATE_PLANES, /* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_CMYK,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};
